//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Core/Computation/IComputation.h
//! @brief     Defines interface IComputation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_CORE_COMPUTATION_ICOMPUTATION_H
#define BORNAGAIN_CORE_COMPUTATION_ICOMPUTATION_H

#include "Core/Computation/ComputationStatus.h"
#include <memory>

class DelayedProgressCounter;
class MultiLayer;
class ProcessedSample;
class ProgressHandler;
class SimulationOptions;

//! Interface for a single-threaded computation with given range of DiffuseElements
//! and ProgressHandler.
//!
//! Controlled by the multi-threading machinery in ISimulation::runSingleSimulation().

class IComputation {
public:
    IComputation(const ProcessedSample& re_sample, const SimulationOptions& options,
                 ProgressHandler& progress);
    virtual ~IComputation();

    //! Calls runProtected(), catches exceptions, sets m_status.
    void run() noexcept;

    bool isCompleted() const { return m_status.isCompleted(); }
    std::string errorMessage() const { return m_status.errorMessage(); }

protected:
    void setProgressHandler(ProgressHandler* progress) const;
    void stepProgress() const;

    const ProcessedSample& m_re_sample;
    const SimulationOptions& m_options;
    ProgressHandler* m_progress;
    ComputationStatus m_status;

private:
    //! Runs computation. May throw. To be called from run(), which catches exceptions.
    virtual void runProtected() = 0;

    mutable std::unique_ptr<DelayedProgressCounter> m_progress_counter;
};

#endif // BORNAGAIN_CORE_COMPUTATION_ICOMPUTATION_H
#endif // USER_API
