//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/AxesItems.cpp
//! @brief     Implements various axis items
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/AxesItems.h"
#include "Base/Axis/FixedBinAxis.h"
#include "GUI/Model/Types/DoubleDescriptor.h"

const int max_detector_pixels = 65536;

BasicAxisItem::BasicAxisItem() : BasicAxisItem(M_TYPE) {}

int BasicAxisItem::binCount() const
{
    return getItemValue(P_NBINS).toInt();
}

void BasicAxisItem::setBinCount(int value)
{
    setItemValue(P_NBINS, value);
}

SessionItem* BasicAxisItem::binsItem() const
{
    return getItem(P_NBINS);
}

DoubleDescriptor BasicAxisItem::lowerBound() const
{
    return DoubleDescriptor(getItem(P_MIN_DEG), Unit::other);
}

void BasicAxisItem::setLowerBound(double value)
{
    setItemValue(P_MIN_DEG, value);
}

SessionItem* BasicAxisItem::lowerBoundItem() const
{
    return getItem(P_MIN_DEG);
}

DoubleDescriptor BasicAxisItem::upperBound() const
{
    return DoubleDescriptor(getItem(P_MAX_DEG), Unit::other);
}

void BasicAxisItem::setUpperBound(double value)
{
    setItemValue(P_MAX_DEG, value);
}

SessionItem* BasicAxisItem::upperBoundItem() const
{
    return getItem(P_MAX_DEG);
}

bool BasicAxisItem::isBoundsPropertiesName(const QString& name)
{
    return name == P_MIN_DEG || name == P_MAX_DEG;
}

QString BasicAxisItem::title() const
{
    return getItemValue(P_TITLE).toString();
}

void BasicAxisItem::setTitle(const QString& title)
{
    setItemValue(P_TITLE, title);
}

SessionItem* BasicAxisItem::titleItem() const
{
    return getItem(P_TITLE);
}

bool BasicAxisItem::isTitlePropertyName(const QString& name)
{
    return name == P_TITLE;
}

bool BasicAxisItem::isTitleVisible() const
{
    return getItemValue(P_TITLE_IS_VISIBLE).toBool();
}

bool BasicAxisItem::isTitleVisiblePropertyName(const QString& name)
{
    return name == P_TITLE_IS_VISIBLE;
}

std::unique_ptr<IAxis> BasicAxisItem::createAxis(double scale) const
{
    return std::make_unique<FixedBinAxis>(title().toStdString(), binCount(), lowerBound() * scale,
                                          upperBound() * scale);
}

SessionItem* BasicAxisItem::visibilityItem() const
{
    return getItem(P_IS_VISIBLE);
}

bool BasicAxisItem::visibilityValue() const
{
    return visibilityItem()->value().toBool();
}

void BasicAxisItem::setVisibilityValue(bool b)
{
    setItemValue(P_IS_VISIBLE, b);
}

bool BasicAxisItem::isVisibilityPropertyName(const QString& name)
{
    return name == P_IS_VISIBLE;
}

BasicAxisItem::~BasicAxisItem() = default;

BasicAxisItem::BasicAxisItem(const QString& type) : SessionItem(type)
{
    register_basic_properties();
}

void BasicAxisItem::register_basic_properties()
{
    addProperty(P_IS_VISIBLE, true)->setVisible(false);
    addProperty(P_NBINS, 100)->setLimits(RealLimits::limited(1, max_detector_pixels));
    addProperty(P_MIN_DEG, 0.0)->setDecimals(3).setLimits(RealLimits::limitless());
    addProperty(P_MAX_DEG, -1.0)->setDecimals(3).setLimits(RealLimits::limitless());
    addProperty(P_TITLE, QString());
    addProperty(P_TITLE_IS_VISIBLE, true)->setVisible(false);
}

// ---------------------------------------------------------------------------------------------- //

AmplitudeAxisItem::AmplitudeAxisItem() : BasicAxisItem(M_TYPE)
{
    addProperty(P_LOCK_MIN_MAX, false)->setVisible(false);
    addProperty(P_IS_LOGSCALE, true);
    visibilityItem()->setVisible(false);
    visibilityItem()->setVisible(true);
    setMinMaxEditor("ScientificDouble");

    mapper()->setOnPropertyChange([this](const QString& name) {
        if (name == P_IS_LOGSCALE) {
            if (getItemValue(P_IS_LOGSCALE).toBool())
                setMinMaxEditor("ScientificDouble");
            else
                setMinMaxEditor("Default");
        }
    });
}

bool AmplitudeAxisItem::isLogScale() const
{
    return getItemValue(P_IS_LOGSCALE).toBool();
}

void AmplitudeAxisItem::setLogScale(bool value)
{
    setItemValue(P_IS_LOGSCALE, value);
}

SessionItem* AmplitudeAxisItem::logScaleItem() const
{
    return getItem(P_IS_LOGSCALE);
}

bool AmplitudeAxisItem::isLogScalePropertyName(const QString& name)
{
    return name == P_IS_LOGSCALE;
}

bool AmplitudeAxisItem::isLocked() const
{
    return getItemValue(P_LOCK_MIN_MAX).toBool();
}

void AmplitudeAxisItem::setLocked(bool locked)
{
    setItemValue(P_LOCK_MIN_MAX, locked);
}


//! Sets editor for min, max values of axes

void AmplitudeAxisItem::setMinMaxEditor(const QString& editorType)
{
    lowerBoundItem()->setEditorType(editorType);
    upperBoundItem()->setEditorType(editorType);
}
