//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/MaskUnitsConverter.cpp
//! @brief     Implements class MaskUnitsConverter
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/MaskUnitsConverter.h"
#include "Device/Data/DataUtils.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/MaskItems.h"
#include "GUI/Model/Data/ProjectionItems.h"

#include <boost/polymorphic_cast.hpp>
using boost::polymorphic_downcast;

MaskUnitsConverter::MaskUnitsConverter() : m_data(nullptr), m_direction(UNDEFINED) {}

//! Converts all masks on board of IntensityDataItem into bin-fraction coordinates.

void MaskUnitsConverter::convertToNbins(IntensityDataItem* intensityData)
{
    m_direction = TO_NBINS;
    convertIntensityDataItem(intensityData);
}

//! Converts all masks on board of IntensityDataItem from bin-fraction coordinates to coordinates
//! of axes currently defined in OutputData.

void MaskUnitsConverter::convertFromNbins(IntensityDataItem* intensityData)
{
    m_direction = FROM_NBINS;
    convertIntensityDataItem(intensityData);
}

//! Converts all masks on board of IntensityDataItem from/to bin-fraction coordinates

void MaskUnitsConverter::convertIntensityDataItem(IntensityDataItem* intensityData)
{
    if (!intensityData || !intensityData->getOutputData())
        return;

    m_data = intensityData->getOutputData();

    if (intensityData->maskContainerItem())
        for (auto* maskItem : intensityData->maskContainerItem()->maskItems())
            convertMask(maskItem);

    if (intensityData->projectionContainerItem())
        for (SessionItem* maskItem : intensityData->projectionContainerItem()->getItems())
            convertMask(maskItem);
}

//! Converts single mask from/to bin-fraction coordinates

void MaskUnitsConverter::convertMask(SessionItem* maskItem)
{
    if (maskItem->modelType() == RectangleItem::M_TYPE
        || maskItem->modelType() == RegionOfInterestItem::M_TYPE) {
        auto* rectItem = polymorphic_downcast<RectangleItem*>(maskItem);
        rectItem->setXLow(convert(rectItem->xLow(), Axis::X));
        rectItem->setYLow(convert(rectItem->yLow(), Axis::Y));
        rectItem->setXUp(convert(rectItem->xUp(), Axis::X));
        rectItem->setYUp(convert(rectItem->yUp(), Axis::Y));
    } else if (maskItem->modelType() == PolygonItem::M_TYPE) {
        for (PolygonPointItem* pointItem : maskItem->childrenOfType<PolygonPointItem>()) {
            pointItem->setPosX(convert(pointItem->posX(), Axis::X));
            pointItem->setPosY(convert(pointItem->posY(), Axis::Y));
        }
    } else if (maskItem->modelType() == VerticalLineItem::M_TYPE) {
        auto* vlineItem = polymorphic_downcast<VerticalLineItem*>(maskItem);
        vlineItem->setPosX(convert(vlineItem->posX(), Axis::X));
    } else if (maskItem->modelType() == HorizontalLineItem::M_TYPE) {
        auto* hlineItem = polymorphic_downcast<HorizontalLineItem*>(maskItem);
        hlineItem->setPosY(convert(hlineItem->posY(), Axis::Y));
    } else if (maskItem->modelType() == EllipseItem::M_TYPE) {
        auto* ellItem = polymorphic_downcast<EllipseItem*>(maskItem);
        double xc = ellItem->xCenter();
        double yc = ellItem->yCenter();
        double xR = ellItem->xRadius();
        double yR = ellItem->yRadius();

        double x2 = xc + xR;
        double y2 = yc + yR;

        if (m_direction == TO_NBINS) {
            DataUtils::Data::coordinateToBinf(xc, yc, *m_data);
            DataUtils::Data::coordinateToBinf(x2, y2, *m_data);
        } else {
            DataUtils::Data::coordinateFromBinf(xc, yc, *m_data);
            DataUtils::Data::coordinateFromBinf(x2, y2, *m_data);
        }
        ellItem->setXCenter(xc);
        ellItem->setYCenter(yc);
        ellItem->setXRadius(x2 - xc);
        ellItem->setYRadius(y2 - yc);
    }
}

//! Convert (x,y) coordinates registered as property (i.e. under P_XPOS, P_YPOS) from/to
//! bin-fraction coordinates. Result of operation are new values for registered properties.

void MaskUnitsConverter::convertCoordinate(SessionItem* maskItem, const QString& xname,
                                           const QString& yname)
{
    if (maskItem->isTag(xname)) {
        double x = convert(maskItem->getItemValue(xname).toDouble(), Axis::X);
        maskItem->setItemValue(xname, x);
    }
    if (maskItem->isTag(yname)) {
        double y = convert(maskItem->getItemValue(yname).toDouble(), Axis::Y);
        maskItem->setItemValue(yname, y);
    }
}

//! Convert value of axis from/to bin-fraction coordinates.

double MaskUnitsConverter::convert(double value, Axis axis)
{
    auto axis_index = static_cast<size_t>(axis);

    ASSERT(m_data);
    ASSERT(axis_index == 0 || axis_index == 1);

    if (m_direction == TO_NBINS)
        return DataUtils::Data::coordinateToBinf(value, m_data->axis(axis_index));
    if (m_direction == FROM_NBINS)
        return DataUtils::Data::coordinateFromBinf(value, m_data->axis(axis_index));
    ASSERT(0);
}
