//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Fit/FitParameterHelper.cpp
//! @brief     Implements namespace FitModelHelper
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Fit/FitParameterHelper.h"
#include "GUI/Model/Fit/FitParameterItem.h"
#include "GUI/Model/Fit/FitParameterContainerItem.h"
#include "GUI/Model/Fit/FitParameterLinkItem.h"
#include "GUI/Model/Fit/ParameterTreeItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Job/JobModel.h"
#include "GUI/Model/Session/ModelPath.h"

//! Creates fit parameter from given ParameterItem, sets starting value to the value
//! of ParameterItem, copies link.

void FitParameterHelper::createFitParameter(FitParameterContainerItem* container,
                                            ParameterItem* parameterItem)
{
    ASSERT(container);
    ASSERT(parameterItem);

    removeFromFitParameters(container, parameterItem);

    auto* model = container->model();
    auto* fitPar = model->insertItem<FitParameterItem>(container);
    fitPar->setDisplayName("par");
    auto* link = model->insertItem<FitParameterLinkItem>(fitPar);
    fitPar->setStartValue(parameterItem->value().toDouble());
    link->setLink(getParameterItemPath(parameterItem));

    fitPar->initMinMaxValues(parameterItem->linkedItem()->limits());
}

//! Removes link to given parameterItem from fit parameters

void FitParameterHelper::removeFromFitParameters(FitParameterContainerItem* container,
                                                 ParameterItem* parameterItem)
{
    FitParameterItem* fitParItem = getFitParameterItem(container, parameterItem);

    if (fitParItem) {
        for (auto* linkItem : fitParItem->linkItems()) {
            if (getParameterItemPath(parameterItem) == linkItem->link()) {
                fitParItem->model()->removeRow(linkItem->index().row(), linkItem->index().parent());
                break;
            }
        }
    }
}

//! Adds given parameterItem to the existing fit parameter with display name fitParName.
//! If parameterItem is already linked with another fitParameter, it will be relinked

void FitParameterHelper::addToFitParameter(FitParameterContainerItem* container,
                                           ParameterItem* parameterItem, const QString& fitParName)
{
    ASSERT(container);

    removeFromFitParameters(container, parameterItem);
    for (FitParameterItem* fitPar : container->fitParameterItems()) {
        if (fitPar->displayName() == fitParName) {
            auto* link = fitPar->model()->insertItem<FitParameterLinkItem>(fitPar);
            link->setLink(getParameterItemPath(parameterItem));
            break;
        }
    }
}

//! Returns fFitParameterItem corresponding to given ParameterItem

FitParameterItem* FitParameterHelper::getFitParameterItem(FitParameterContainerItem* container,
                                                          ParameterItem* parameterItem)
{
    ASSERT(container);
    return container->fitParameterItem(getParameterItemPath(parameterItem));
}

//! Returns list of fit parameter display names

QStringList FitParameterHelper::getFitParameterNames(FitParameterContainerItem* container)
{
    ASSERT(container);
    QStringList result;
    for (FitParameterItem* item : container->fitParameterItems()) {
        result.append(item->displayName());
    }
    return result;
}

//! return path to given item in the ParameterTreeContainer

QString FitParameterHelper::getParameterItemPath(const ParameterItem* parameterItem)
{
    QString result = GUI::Model::Path::getPathFromIndex(parameterItem->index());
    QString containerPrefix = "Parameter Container/";
    int containerEnd = result.indexOf(containerPrefix) + containerPrefix.size();
    result = result.mid(containerEnd);
    return result;
}

//! Returns ParameterItem corresponding to given link.
//! Link is relative to ParameterContainerItem, so first we have to find it

ParameterItem* FitParameterHelper::getParameterItem(FitParameterContainerItem* container,
                                                    const QString& link)
{
    SessionItem* cur = container;
    while (cur && cur->modelType() != JobItem::M_TYPE) {
        cur = cur->parent();
    }
    ASSERT(cur && cur->modelType() == JobItem::M_TYPE);
    auto* jobItem = dynamic_cast<JobItem*>(cur);
    ASSERT(jobItem);
    return dynamic_cast<ParameterItem*>(
        GUI::Model::Path::getItemFromPath(link, jobItem->parameterContainerItem()));
}
