//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Group/DistributionItems.h
//! @brief     Defines class DistributionItem and several subclasses
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_GROUP_DISTRIBUTIONITEMS_H
#define BORNAGAIN_GUI_MODEL_GROUP_DISTRIBUTIONITEMS_H

#include "GUI/Model/Session/SessionItem.h"

class IDistribution1D;
class IRangedDistribution;
class RealLimitsItem;

class BA_CORE_API_ DistributionItem : public SessionItem {
private:
    static constexpr auto P_NUMBER_OF_SAMPLES{"Number of samples"};
    static constexpr auto P_SIGMA_FACTOR{"Sigma factor"};
    static constexpr auto P_IS_INITIALIZED{"is initialized"};
    static constexpr auto P_LIMITS{"Limits"};

public:
    explicit DistributionItem(const QString& name);

    virtual std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const = 0;
    void init_parameters(double value, const RealLimits& limits = RealLimits::limitless());
    void init_limits_group(const RealLimits& limits, double factor = 1.0);
    virtual void showMean(bool) = 0;

    size_t numberOfSamples() const;
    void setNumberOfSamples(size_t number);

    bool hasSigmaFactor() const;
    double sigmaFactor() const;
    void setSigmaFactor(double sigma_factor);

    bool hasLimits() const;
    SessionItem* limitsItem() const;
    RealLimitsItem* limits() const;

protected:
    virtual void init_distribution(double) {}
    void register_number_of_samples();
    void register_sigma_factor();
    void register_limits();
};

class SymmetricDistributionItem : public DistributionItem {
private:
    static constexpr auto P_MEAN{"Mean"};

public:
    SymmetricDistributionItem(const QString& name, double mean);
    void showMean(bool flag) override;

    SessionItem* meanItem() const;
    double mean() const;
    void setMean(double mean);
    static bool isMeanPropertyName(const QString& name);

    virtual std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const = 0;
    virtual double deviation(double scale) const = 0;
};

class DistributionNoneItem : public SymmetricDistributionItem {

public:
    static constexpr auto M_TYPE{"DistributionNone"};

    DistributionNoneItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;
    void init_distribution(double value) override;
};

class DistributionGateItem : public DistributionItem {
private:
    static constexpr auto P_MIN{"Min"};
    static constexpr auto P_MAX{"Max"};

public:
    static constexpr auto M_TYPE{"DistributionGate"};

    DistributionGateItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void init_distribution(double value) override;
    void showMean(bool) override {}

    double minimum() const;
    double maximum() const;
    void setRange(double min, double max);
};

class DistributionLorentzItem : public SymmetricDistributionItem {
private:
    static constexpr auto P_HWHM{"HWHM"};

public:
    static constexpr auto M_TYPE{"DistributionLorentz"};

    DistributionLorentzItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;

    double hwhm() const;
    void setHWHM(double hwhm);

    void init_distribution(double value) override;
};

class DistributionGaussianItem : public SymmetricDistributionItem {
private:
    static constexpr auto P_STD_DEV{"StdDev"};

public:
    static constexpr auto M_TYPE{"DistributionGaussian"};

    DistributionGaussianItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;

    double standardDeviation() const;
    void setStandardDeviation(double std_dev);
    void init_distribution(double value) override;
};

class DistributionLogNormalItem : public DistributionItem {
private:
    static constexpr auto P_MEDIAN{"Median"};
    static constexpr auto P_SCALE_PAR{"ScaleParameter"};

public:
    static constexpr auto M_TYPE{"DistributionLogNormal"};

    DistributionLogNormalItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void init_distribution(double value) override;
    void showMean(bool flag) override;

    double median() const;
    void setMedian(double median);

    double scaleParameter() const;
    void setScaleParameter(double scale_par);
};

class DistributionCosineItem : public SymmetricDistributionItem {
private:
    static constexpr auto P_SIGMA{"Sigma"};

public:
    static constexpr auto M_TYPE{"DistributionCosine"};

    DistributionCosineItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    std::unique_ptr<IRangedDistribution> createIRangedDistribution(double scale) const override;
    double deviation(double scale) const override;
    void init_distribution(double value) override;

    double sigma() const;
    void setSigma(double sigma);
};

class DistributionTrapezoidItem : public DistributionItem {
private:
    static constexpr auto P_CENTER{"Center"};
    static constexpr auto P_LEFTWIDTH{"LeftWidth"};
    static constexpr auto P_MIDDLEWIDTH{"MiddleWidth"};
    static constexpr auto P_RIGHTWIDTH{"RightWidth"};

public:
    static constexpr auto M_TYPE{"DistributionTrapezoid"};

    DistributionTrapezoidItem();

    std::unique_ptr<IDistribution1D> createDistribution(double scale = 1.0) const override;
    void init_distribution(double value) override;
    void showMean(bool flag) override;

    double center() const;
    void setCenter(double center);

    double leftWidth() const;
    void setLeftWidth(double left_width);

    double middleWidth() const;
    void setMiddleWidth(double middle_width);

    double rightWidth() const;
    void setRightWidth(double right_width);
};

#endif // BORNAGAIN_GUI_MODEL_GROUP_DISTRIBUTIONITEMS_H
