//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/BeamDistributionItem.cpp
//! @brief     Implements class BeamDistributionItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Instrument/BeamDistributionItem.h"
#include "GUI/Model/Group/GroupItem.h"
#include "GUI/Model/Session/RealLimitsItems.h"
#include "Param/Distrib/Distributions.h"

namespace {

// create a GroupInfo containing all distributions
GroupInfo createAllDistributionsGroup()
{
    GroupInfo info;
    info.add(DistributionNoneItem::M_TYPE, "None");
    info.add(DistributionGateItem::M_TYPE, "Gate");
    info.add(DistributionLorentzItem::M_TYPE, "Lorentz");
    info.add(DistributionGaussianItem::M_TYPE, "Gaussian");
    info.add(DistributionLogNormalItem::M_TYPE, "Log Normal");
    info.add(DistributionCosineItem::M_TYPE, "Cosine");
    info.add(DistributionTrapezoidItem::M_TYPE, "Trapezoid");
    info.setDefaultType(DistributionNoneItem::M_TYPE);
    return info;
}

// create a GroupInfo containing all symmetric distributions
GroupInfo createSymmetricDistributionsGroup()
{
    GroupInfo info;
    info.add(DistributionNoneItem::M_TYPE, "None");
    info.add(DistributionLorentzItem::M_TYPE, "Lorentz");
    info.add(DistributionGaussianItem::M_TYPE, "Gaussian");
    info.add(DistributionCosineItem::M_TYPE, "Cosine");
    info.setDefaultType(DistributionNoneItem::M_TYPE);
    return info;
}

} // namespace

BeamDistributionItem::BeamDistributionItem(const QString& name, bool show_mean) : SessionItem(name)
{
    mapper()->setOnChildPropertyChange([this, show_mean](SessionItem* item, const QString&) {
        if (item->hasModelType<GroupItem>() && item->parent() == this)
            initDistributionItem(show_mean);
    });
}

//! returns parameter distribution to add into the ISimulation
std::unique_ptr<ParameterDistribution> BeamDistributionItem::getParameterDistributionForName(
    ParameterDistribution::WhichParameter which) const
{
    std::unique_ptr<ParameterDistribution> P_par_distr{};
    if (auto* distributionItem = dynamic_cast<DistributionItem*>(getGroupItem(P_DISTRIBUTION))) {
        auto P_distribution = createDistribution1D();

        if (P_distribution) {
            size_t nbr_samples = distributionItem->numberOfSamples();
            double sigma_factor(0);
            if (distributionItem->hasSigmaFactor()) {
                sigma_factor = distributionItem->sigmaFactor();
            }

            RealLimitsItem* limitsItem = distributionItem->limits();
            ASSERT(limitsItem);

            RealLimits limits = limitsItem->createRealLimits(scaleFactor());

            P_par_distr = std::make_unique<ParameterDistribution>(
                which, *P_distribution, nbr_samples, sigma_factor, limits);
        }
    }
    return P_par_distr;
}

//! Propagates the value and limits stored in DistributionNone type into alls distributions.

void BeamDistributionItem::initDistributionItem(bool show_mean)
{
    auto* groupItem = item<GroupItem>(P_DISTRIBUTION);

    DistributionNoneItem* distributionNone = nullptr;
    for (auto* item : groupItem->groupItems()) {
        if (item->modelType() == DistributionNoneItem::M_TYPE) {
            distributionNone = dynamic_cast<DistributionNoneItem*>(item);
            break;
        }
    }

    if (!distributionNone)
        return;

    const RealLimits limits = distributionNone->meanItem()->limits();
    const QString editor_type = distributionNone->meanItem()->editorType();

    for (auto* item : groupItem->groupItems()) {
        auto* distrItem = dynamic_cast<DistributionItem*>(item);
        distrItem->showMean(show_mean);

        if (item == distributionNone)
            continue;

        distrItem->init_parameters(distributionNone->mean(), limits);
        if (auto* symmetric_distr = dynamic_cast<SymmetricDistributionItem*>(distrItem))
            symmetric_distr->meanItem()->setEditorType(editor_type).setLimits(limits);

        // hiding limits from the editor
        if (distrItem->hasLimits())
            distrItem->limitsItem()->setVisible(false);
    }
}

//! Returns mean value of the distribution.

double BeamDistributionItem::meanValue() const
{
    std::unique_ptr<IDistribution1D> domainDistr = createDistribution1D();
    if (domainDistr)
        return domainDistr->getMean() / scaleFactor();
    return dynamic_cast<SymmetricDistributionItem*>(getGroupItem(P_DISTRIBUTION))->mean();
}

void BeamDistributionItem::resetToValue(double value)
{
    auto* distributionItem = dynamic_cast<DistributionNoneItem*>(
        setGroupProperty(P_DISTRIBUTION, DistributionNoneItem::M_TYPE));
    ASSERT(distributionItem);
    distributionItem->setMean(value);
}

//! Scales the values provided by distribution (to perform deg->rad conversion in the case
//! of AngleDistributionItems.

double BeamDistributionItem::scaleFactor() const
{
    return 1.0;
}

DistributionItem* BeamDistributionItem::distribution() const
{
    return dynamic_cast<DistributionItem*>(getGroupItem(P_DISTRIBUTION));
}

DistributionItem* BeamDistributionItem::setDistributionType(const QString& model_type)
{
    ASSERT(model_type == DistributionGateItem::M_TYPE
           || model_type == DistributionLogNormalItem::M_TYPE
           || model_type == DistributionTrapezoidItem::M_TYPE
           || model_type == DistributionCosineItem::M_TYPE
           || model_type == DistributionGaussianItem::M_TYPE
           || model_type == DistributionLorentzItem::M_TYPE
           || model_type == DistributionNoneItem::M_TYPE);

    return dynamic_cast<DistributionItem*>(setGroupProperty(P_DISTRIBUTION, model_type));
}

SessionItem* BeamDistributionItem::distributionItem() const
{
    return getItem(P_DISTRIBUTION);
}

void BeamDistributionItem::register_distribution_group(const Distributions distributions)
{
    if (distributions == Distributions::All) {
        static const GroupInfo all = createAllDistributionsGroup();
        addGroupProperty(P_DISTRIBUTION, all);
    } else if (distributions == Distributions::Symmetric) {
        static const GroupInfo symmetric = createSymmetricDistributionsGroup();
        addGroupProperty(P_DISTRIBUTION, symmetric);
    } else
        ASSERT(false);
}

std::unique_ptr<IDistribution1D> BeamDistributionItem::createDistribution1D() const
{
    if (auto* distItem = dynamic_cast<DistributionItem*>(getGroupItem(P_DISTRIBUTION)))
        return distItem->createDistribution(scaleFactor());

    return {};
}
