//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/InstrumentLibrary.cpp
//! @brief     Implements class InstrumentLibrary
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Instrument/InstrumentLibrary.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Instrument/InstrumentsTreeModel.h"
#include "GUI/Model/State/SessionData.h"
#include <QFile>
#include <QXmlStreamWriter>

namespace {

const QString XML_ROOT_TAG = "BornAgainInstrumentLibrary";
const QString XML_VERSION_TAG = "Version";

QString instrumentLibraryFilePath()
{
    return SessionData::appDataFolder() + "/BornAgainInstrumentLibrary.balib";
}

} // namespace


InstrumentLibrary::InstrumentLibrary() : m_modified(false)
{
    QObject::connect(&m_model, &InstrumentModel::instrumentChanged, [&]() { m_modified = true; });
}

QString InstrumentLibrary::suggestName(const QString& name) const
{
    return m_model.suggestInstrumentName(name);
}

InstrumentItem* InstrumentLibrary::add(const QString& name, const InstrumentItem& itemToCopy)
{
    auto* p = m_model.insertCopy(itemToCopy);
    p->setInstrumentName(name);
    m_modified = true;
    return p;
}

bool InstrumentLibrary::isEmpty() const
{
    return m_model.instrumentItems().isEmpty();
}

QList<InstrumentItem*> InstrumentLibrary::instrumentItems() const
{
    return m_model.instrumentItems().toList();
}

bool InstrumentLibrary::saveIfModified()
{
    if (!m_modified)
        return true;

    QFile file(instrumentLibraryFilePath());
    if (!file.open(QFile::ReadWrite | QIODevice::Truncate | QFile::Text))
        return false;

    QXmlStreamWriter writer(&file);
    writer.setAutoFormatting(true);
    writer.writeStartDocument();
    writer.writeStartElement(XML_ROOT_TAG);
    writer.writeAttribute(XML_VERSION_TAG, "1");

    m_model.writeTo(&writer);

    writer.writeEndElement();
    writer.writeEndDocument();

    file.close();
    m_modified = false;
    return true;
}

bool InstrumentLibrary::load()
{
    m_modified = false;
    m_model.clear();

    QFile file(instrumentLibraryFilePath());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return false;

    try {
        QXmlStreamReader reader(&file);

        while (!reader.atEnd()) {
            reader.readNext();
            if (reader.isStartElement()) {
                if (reader.name() == XML_ROOT_TAG) {
                    int foundVersion =
                        reader.attributes().value(XML_VERSION_TAG).toString().toInt();
                    if (foundVersion != 1)
                        return false;

                    reader.readNextStartElement();
                    if (m_model.getModelTag() == reader.name())
                        m_model.readFrom(&reader);
                }
            }
        }

        if (reader.hasError()) {
            m_model.clear();
            return false;
        }

        file.close();
        return true;
    } catch (const std::exception&) {
        m_model.clear();
        return false;
    }
}

InstrumentModel* InstrumentLibrary::model()
{
    return &m_model;
}
