//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/JobItemUtils.cpp
//! @brief     Implements class JobItemUtils
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/JobItemUtils.h"
#include "Core/Simulation/ISimulation.h"
#include "Device/Histo/SimulationResult.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Instrument/InstrumentItems.h"

namespace {

const std::map<QString, Axes::Coords> units_from_names{{"nbins", Axes::Coords::NBINS},
                                                       {"Radians", Axes::Coords::RADIANS},
                                                       {"Degrees", Axes::Coords::DEGREES},
                                                       {"mm", Axes::Coords::MM},
                                                       {"q-space", Axes::Coords::QSPACE}};

const std::map<Axes::Coords, QString> names_from_coords{{Axes::Coords::NBINS, "nbins"},
                                                        {Axes::Coords::RADIANS, "Radians"},
                                                        {Axes::Coords::MM, "mm"},
                                                        {Axes::Coords::QSPACE, "q-space"},
                                                        {Axes::Coords::DEGREES, "Degrees"}};

//! Updates axes' titles
void updateAxesTitle(DataItem* intensityItem, const ICoordSystem& converter, Axes::Coords units)
{
    intensityItem->setXaxisTitle(QString::fromStdString(converter.axisName(0, units)));
    if (converter.dimension() > 1)
        intensityItem->setYaxisTitle(QString::fromStdString(converter.axisName(1, units)));
}

} // namespace


//! Updates axes of OutputData in IntensityData item to correspond with ::P_AXES_UNITS selection.
//! InstrumentItem is used to get domain's detector map for given units.

void GUI::Model::JobItemUtils::updateDataAxes(DataItem* intensityItem,
                                              const InstrumentItem* instrumentItem)
{
    ASSERT(intensityItem);

    if (!instrumentItem) {
        // special case when reading old project files: project failed on load instrument
        // but we want to try to recover the rest.
        return;
    }

    if (!intensityItem->getOutputData())
        return;

    Axes::Coords requested_units = coordsFromName(intensityItem->selectedCoords());

    auto* const converter = instrumentItem->createCoordSystem();
    auto newData = converter->createOutputData(requested_units);
    newData->setRawDataVector(intensityItem->getOutputData()->getRawDataVector());

    intensityItem->setOutputData(newData.release());
    intensityItem->setAxesRangeToData();
    updateAxesTitle(intensityItem, *converter, requested_units);
}

//! Correspondence of domain detector axes types to their GUI counterpart.

QString GUI::Model::JobItemUtils::nameFromCoords(Axes::Coords units)
{
    const auto it = names_from_coords.find(units);
    if (it == names_from_coords.end())
        return "";
    return it->second;
}

//! Correspondence of GUI axes units names to their domain counterpart.

Axes::Coords GUI::Model::JobItemUtils::coordsFromName(const QString& name)
{
    return units_from_names.at(name);
}

//! Sets axes units suitable for given instrument.

void GUI::Model::JobItemUtils::setIntensityItemCoords(DataItem* intensityItem,
                                                      const InstrumentItem* instrumentItem)
{
    auto* const converter = instrumentItem->createCoordSystem();
    if (!converter)
        return;
    setIntensityItemCoords(intensityItem, *converter);
}

void GUI::Model::JobItemUtils::setIntensityItemCoords(DataItem* intensityItem,
                                                      const ICoordSystem& converter)
{
    ComboProperty combo = availableUnits(converter);
    intensityItem->setAxesUnits(combo);
}

void GUI::Model::JobItemUtils::createDefaultDetectorMap(DataItem* intensityItem,
                                                        const Instrument2DItem* instrumentItem)
{
    auto* const converter = instrumentItem->createCoordSystem();
    auto output_data = converter->createOutputData(converter->defaultUnits());
    intensityItem->setOutputData(output_data.release());
    setIntensityItemCoords(intensityItem, *converter);
    updateAxesTitle(intensityItem, *converter, converter->defaultUnits());
}

void GUI::Model::JobItemUtils::setResults(DataItem* intensityItem, const ISimulation* simulation)
{
    const auto sim_result = simulation->result();
    if (intensityItem->getOutputData() == nullptr) {
        const auto& converter = sim_result.converter();
        GUI::Model::JobItemUtils::setIntensityItemCoords(intensityItem, converter);
        updateAxesTitle(intensityItem, converter, converter.defaultUnits());
    }
    auto selected_units = GUI::Model::JobItemUtils::coordsFromName(intensityItem->selectedCoords());
    auto data = sim_result.data(selected_units);
    intensityItem->setOutputData(data.release());
}

ComboProperty GUI::Model::JobItemUtils::availableUnits(const ICoordSystem& converter)
{
    ComboProperty result;
    for (auto units : converter.availableUnits()) {
        auto unit_name = nameFromCoords(units);
        if (!unit_name.isEmpty())
            result << unit_name;
    }

    result.setValue(nameFromCoords(converter.defaultUnits()));
    return result;
}
