//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Material/MaterialItem.cpp
//! @brief     Implements class MaterialItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Material/MaterialItem.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Group/GroupInfo.h"
#include "GUI/Model/Material/MaterialDataItems.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "Sample/Material/MaterialFactoryFuncs.h"
#include <QColor>
#include <QUuid>

namespace {

const QString magnetization_tooltip = "Magnetization (A/m)";

}


MaterialItem::MaterialItem() : SessionItem(M_TYPE)
{
    setItemName("Material");

    addProperty(P_COLOR, QColor(Qt::red));

    GroupInfo info;
    info.add(MaterialRefractiveDataItem::M_TYPE, "Refractive index based");
    info.add(MaterialSLDDataItem::M_TYPE, "SLD based");
    info.setDefaultType(MaterialRefractiveDataItem::M_TYPE);
    addGroupProperty(P_MATERIAL_DATA, info);

    addProperty<VectorItem>(P_MAGNETIZATION)->setToolTip(magnetization_tooltip);
    addProperty(P_IDENTIFIER, QUuid::createUuid().toString());
    getItem(P_IDENTIFIER)->setVisible(false);
}

//! Turns material into refractive index material.

void MaterialItem::setRefractiveIndex(const double delta, const double beta)
{
    auto* refractiveData = setGroupProperty(P_MATERIAL_DATA, MaterialRefractiveDataItem::M_TYPE);
    refractiveData->setItemValue(MaterialRefractiveDataItem::P_DELTA, delta);
    refractiveData->setItemValue(MaterialRefractiveDataItem::P_BETA, beta);
}

//! Turns material into SLD based material.

void MaterialItem::setScatteringLengthDensity(const complex_t sld)
{
    auto* sldData = setGroupProperty(P_MATERIAL_DATA, MaterialSLDDataItem::M_TYPE);
    sldData->setItemValue(MaterialSLDDataItem::P_SLD_REAL, sld.real());
    sldData->setItemValue(MaterialSLDDataItem::P_SLD_IMAG, sld.imag());
}

DoubleDescriptor MaterialItem::delta()
{
    return DoubleDescriptor(
        getGroupItem(P_MATERIAL_DATA)->getItem(MaterialRefractiveDataItem::P_DELTA),
        Unit::unitless);
}

DoubleDescriptor MaterialItem::beta()
{
    return DoubleDescriptor(
        getGroupItem(P_MATERIAL_DATA)->getItem(MaterialRefractiveDataItem::P_BETA), Unit::unitless);
}

DoubleDescriptor MaterialItem::sldRe()
{
    return DoubleDescriptor(getGroupItem(P_MATERIAL_DATA)->getItem(MaterialSLDDataItem::P_SLD_REAL),
                            Unit::unitless);
}

DoubleDescriptor MaterialItem::sldIm()
{
    return DoubleDescriptor(getGroupItem(P_MATERIAL_DATA)->getItem(MaterialSLDDataItem::P_SLD_IMAG),
                            Unit::unitless);
}

VectorDescriptor MaterialItem::magnetizationVector()
{
    return VectorDescriptor(item<VectorItem>(P_MAGNETIZATION), "A/m");
}

bool MaterialItem::hasRefractiveIndex() const
{
    SessionItem* data = getGroupItem(P_MATERIAL_DATA);
    ASSERT(data);
    return dynamic_cast<MaterialRefractiveDataItem*>(data) != nullptr;
}

double MaterialItem::refractiveIndexDelta() const
{
    ASSERT(hasRefractiveIndex());
    return getGroupItem(P_MATERIAL_DATA)
        ->getItemValue(MaterialRefractiveDataItem::P_DELTA)
        .toDouble();
}

double MaterialItem::refractiveIndexBeta() const
{
    ASSERT(hasRefractiveIndex());
    return getGroupItem(P_MATERIAL_DATA)
        ->getItemValue(MaterialRefractiveDataItem::P_BETA)
        .toDouble();
}

complex_t MaterialItem::scatteringLengthDensity() const
{
    ASSERT(!hasRefractiveIndex());
    SessionItem* ri = getGroupItem(P_MATERIAL_DATA);
    double real = ri->getItemValue(MaterialSLDDataItem::P_SLD_REAL).toDouble();
    double imag = ri->getItemValue(MaterialSLDDataItem::P_SLD_IMAG).toDouble();
    return complex_t(real, imag);
}

QString MaterialItem::materialName() const
{
    return itemName();
}

void MaterialItem::setMaterialName(const QString& name)
{
    setItemName(name);
}

QString MaterialItem::identifier() const
{
    return getItemValue(P_IDENTIFIER).toString();
}

void MaterialItem::setIdentifier(const QString& id)
{
    setItemValue(P_IDENTIFIER, id);
}

QColor MaterialItem::color() const
{
    return getItemValue(P_COLOR).value<QColor>();
}

void MaterialItem::setColor(const QColor& color)
{
    setItemValue(P_COLOR, color);
}

R3 MaterialItem::magnetization() const
{
    return item<VectorItem>(P_MAGNETIZATION)->getVector();
}

void MaterialItem::setMagnetization(const R3& magnetization)
{
    item<VectorItem>(P_MAGNETIZATION)->setVector(magnetization);
}

void MaterialItem::hideMagnetization()
{
    getItem(MaterialItem::P_MAGNETIZATION)->setVisible(false);
}

std::unique_ptr<Material> MaterialItem::createMaterial() const
{
    auto* dataItem = getGroupItem(P_MATERIAL_DATA);
    auto magnetization = item<VectorItem>(P_MAGNETIZATION)->getVector();
    auto name = itemName().toStdString();

    if (dataItem->modelType() == MaterialRefractiveDataItem::M_TYPE) {
        double delta = dataItem->getItemValue(MaterialRefractiveDataItem::P_DELTA).toDouble();
        double beta = dataItem->getItemValue(MaterialRefractiveDataItem::P_BETA).toDouble();
        return std::make_unique<Material>(HomogeneousMaterial(name, delta, beta, magnetization));
    }
    if (dataItem->modelType() == MaterialSLDDataItem::M_TYPE) {
        double sld_real = dataItem->getItemValue(MaterialSLDDataItem::P_SLD_REAL).toDouble();
        double sld_imag = dataItem->getItemValue(MaterialSLDDataItem::P_SLD_IMAG).toDouble();
        return std::make_unique<Material>(MaterialBySLD(name, sld_real, sld_imag, magnetization));
    }
    ASSERT(0);
}

bool MaterialItem::operator!=(const MaterialItem& other) const
{
    return !operator==(other);
}

bool MaterialItem::operator==(const MaterialItem& other) const
{
    if (identifier() != other.identifier())
        return false;
    if (itemName() != other.itemName())
        return false;
    if (color() != other.color())
        return false;
    if (magnetization() != other.magnetization())
        return false;

    if (hasRefractiveIndex() != other.hasRefractiveIndex())
        return false;

    if (hasRefractiveIndex()) {
        if (refractiveIndexBeta() != other.refractiveIndexBeta())
            return false;
        if (refractiveIndexDelta() != other.refractiveIndexDelta())
            return false;
    } else {
        if (scatteringLengthDensity() != other.scatteringLengthDensity())
            return false;
    }

    return true;
}
