//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleCompositionItem.cpp
//! @brief     Implements class ParticleCompositionItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/VectorItem.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace


ParticleCompositionItem::ParticleCompositionItem()
    : SessionItem(M_TYPE), ItemWithParticles(M_TYPE, abundance_tooltip, position_tooltip)
{
    setToolTip("Composition of particles with fixed positions");

    registerTag(T_PARTICLES, 0, -1,
                {ParticleItem::M_TYPE, ParticleCoreShellItem::M_TYPE,
                 ParticleCompositionItem::M_TYPE, MesoCrystalItem::M_TYPE});
    setDefaultTag(T_PARTICLES);
}

std::unique_ptr<ParticleComposition> ParticleCompositionItem::createParticleComposition() const
{
    auto P_composition = std::make_unique<ParticleComposition>();
    P_composition->setAbundance(abundance());
    QVector<SessionItem*> childlist = children();
    for (int i = 0; i < childlist.size(); ++i) {
        if (childlist[i]->hasModelType<ParticleItem>()) {
            auto* particle_item = dynamic_cast<ParticleItem*>(childlist[i]);
            auto P_particle = particle_item->createParticle();
            if (P_particle) {
                P_composition->addParticle(*P_particle);
            }
        } else if (childlist[i]->hasModelType<ParticleCoreShellItem>()) {
            auto* particle_coreshell_item = dynamic_cast<ParticleCoreShellItem*>(childlist[i]);
            auto P_particle_coreshell = particle_coreshell_item->createParticleCoreShell();
            if (P_particle_coreshell) {
                P_composition->addParticle(*P_particle_coreshell);
            }
        } else if (childlist[i]->hasModelType<ParticleCompositionItem>()) {
            auto* particlecomposition_item = dynamic_cast<ParticleCompositionItem*>(childlist[i]);
            auto P_child_composition = particlecomposition_item->createParticleComposition();
            if (P_child_composition) {
                P_composition->addParticle(*P_child_composition);
            }
        } else if (childlist[i]->hasModelType<MesoCrystalItem>()) {
            auto* mesocrystal_item = dynamic_cast<MesoCrystalItem*>(childlist[i]);
            auto P_child_meso = mesocrystal_item->createMesoCrystal();
            if (P_child_meso) {
                P_composition->addParticle(*P_child_meso);
            }
        }
    }
    setTransformationInfo(P_composition.get());

    return P_composition;
}

QVector<ItemWithParticles*> ParticleCompositionItem::particles() const
{
    return items<ItemWithParticles>(T_PARTICLES);
}

void ParticleCompositionItem::addParticle(ItemWithParticles* particle)
{
    model()->moveItem(particle, this, -1, T_PARTICLES);
}

void ParticleCompositionItem::removeParticle(ItemWithParticles* particle)
{
    model()->removeItem(particle);
}
