//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleItem.cpp
//! @brief     Implements class ParticleItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Group/GroupItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Job/JobModelFunctions.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Material/MaterialItemUtils.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/Types/VectorItem.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/IFormFactor.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace

ParticleItem::ParticleItem()
    : SessionItem(M_TYPE)
    , ItemWithMaterial(M_TYPE)
    , ItemWithParticles(M_TYPE, abundance_tooltip, position_tooltip)
{
    addGroupProperty(P_FORM_FACTOR, formFactorGroupInfo());

    setDefaultTagTransformation();
}

std::unique_ptr<Particle> ParticleItem::createParticle() const
{
    auto& ffItem = groupItem<FormFactorItem>(P_FORM_FACTOR);
    auto domainMaterial = materialItem()->createMaterial();

    auto particle = std::make_unique<Particle>(*domainMaterial, *ffItem.createFormFactor());
    particle->setAbundance(abundance());

    setTransformationInfo(particle.get());

    return particle;
}

SessionItem* ParticleItem::setFormFactor(const QString& model_type)
{
    return setGroupProperty(P_FORM_FACTOR, model_type);
}

bool ParticleItem::isFormFactorPropertyName(const QString& name)
{
    return name == P_FORM_FACTOR;
}

FormFactorItem* ParticleItem::formFactor() const
{
    return dynamic_cast<FormFactorItem*>(getGroupItem(P_FORM_FACTOR));
}

GroupItem* ParticleItem::formFactorItem() const
{
    return item<GroupItem>(P_FORM_FACTOR);
}

//! Returns true if this particle is a shell particle.

bool ParticleItem::isShellParticle() const
{
    if (!parent())
        return false;

    return parent()->hasModelType<ParticleCoreShellItem>()
           && ParticleCoreShellItem::isShellTagName(parent()->tagFromItem(this));
}

//! Returns true if this particle is directly connected to a ParticleLayout

bool ParticleItem::parentIsParticleLayout() const
{
    if (!parent())
        return false;

    return parent()->hasModelType<ParticleLayoutItem>();
}
