//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Session/RealLimitsItems.cpp
//! @brief     Defines RealLimitsItems's classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Session/RealLimitsItems.h"

namespace {
const QString tooltip_min_value = "Minimum allowed value, value included.";
const QString tooltip_max_value = "Maximum allowed value, value excluded.";
} // namespace

RealLimitsItem::RealLimitsItem(const QString& name, const std::optional<double>& lower,
                               const std::optional<double>& upper)
    : SessionItem(name)
{
    if (lower)
        addProperty(P_XMIN, lower.value())
            ->setToolTip(tooltip_min_value)
            .setLimits(RealLimits::limitless());

    if (upper)
        addProperty(P_XMAX, upper.value())
            ->setToolTip(tooltip_max_value)
            .setLimits(RealLimits::limitless());
}

double RealLimitsItem::lowerLimit() const
{
    return getItemValue(P_XMIN).toDouble();
}

void RealLimitsItem::setLowerLimit(double lower)
{
    setItemValue(P_XMIN, lower);
}

double RealLimitsItem::upperLimit() const
{
    return getItemValue(P_XMAX).toDouble();
}

void RealLimitsItem::setUpperLimit(double upper)
{
    setItemValue(P_XMAX, upper);
}

// --------------------------------------------------------------------------------------------- //

LimitlessItem::LimitlessItem() : RealLimitsItem(M_TYPE) {}

RealLimits LimitlessItem::createRealLimits(double /*scale_factor*/) const
{
    return RealLimits();
}

// --------------------------------------------------------------------------------------------- //

PositiveItem::PositiveItem() : RealLimitsItem(M_TYPE) {}

RealLimits PositiveItem::createRealLimits(double /*scale_factor*/) const
{
    return RealLimits::positive();
}

// --------------------------------------------------------------------------------------------- //

NonnegativeItem::NonnegativeItem() : RealLimitsItem(M_TYPE) {}

RealLimits NonnegativeItem::createRealLimits(double /*scale_factor*/) const
{
    return RealLimits::nonnegative();
}

// --------------------------------------------------------------------------------------------- //

LowerLimitedItem::LowerLimitedItem() : RealLimitsItem(M_TYPE, 0.0) {}

RealLimits LowerLimitedItem::createRealLimits(double scale_factor) const
{
    return RealLimits::lowerLimited(scale_factor * lowerLimit());
}

// --------------------------------------------------------------------------------------------- //

UpperLimitedItem::UpperLimitedItem() : RealLimitsItem(M_TYPE, std::nullopt, 1.0) {}

RealLimits UpperLimitedItem::createRealLimits(double scale_factor) const
{
    return RealLimits::upperLimited(scale_factor * upperLimit());
}

// --------------------------------------------------------------------------------------------- //

LimitedItem::LimitedItem() : RealLimitsItem(M_TYPE, 0.0, 1.0) {}

RealLimits LimitedItem::createRealLimits(double scale_factor) const
{
    return RealLimits::limited(scale_factor * lowerLimit(), scale_factor * upperLimit());
}
