//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/To/ToDomain.cpp
//! @brief     Implements class TransformToDomain
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/To/ToDomain.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/LayerRoughnessItems.h"
#include "GUI/Model/Sample/MesoCrystalItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/UIntDescriptor.h"
#include "Sample/Particle/MesoCrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"


std::unique_ptr<MultiLayer> GUI::Transform::ToDomain::createMultiLayer(const MultiLayerItem& item)
{
    auto P_multilayer = std::make_unique<MultiLayer>();
    double cross_corr_length = item.crossCorrLength();
    if (cross_corr_length > 0)
        P_multilayer->setCrossCorrLength(cross_corr_length);
    R3 external_field = item.externalField();
    P_multilayer->setExternalField(external_field);
    return P_multilayer;
}

std::unique_ptr<Layer> GUI::Transform::ToDomain::createLayer(const LayerItem& item)
{
    auto P_layer =
        std::make_unique<Layer>(*item.materialItem()->createMaterial(), item.thickness());
    P_layer->setNumberOfSlices(item.numSlices());
    return P_layer;
}

std::unique_ptr<LayerRoughness> GUI::Transform::ToDomain::createLayerRoughness(
    const std::variant<LayerZeroRoughnessItem*, LayerBasicRoughnessItem*>& roughness)
{
    if (std::holds_alternative<LayerZeroRoughnessItem*>(roughness))
        return nullptr;
    if (std::holds_alternative<LayerBasicRoughnessItem*>(roughness)) {
        const auto& basicRoughnessItem = *std::get<LayerBasicRoughnessItem*>(roughness);
        return std::make_unique<LayerRoughness>(basicRoughnessItem.sigma(),
                                                basicRoughnessItem.hurst(),
                                                basicRoughnessItem.lateralCorrelationLength());
    }
    ASSERT(0);
}

std::unique_ptr<ParticleLayout>
GUI::Transform::ToDomain::createParticleLayout(const ParticleLayoutItem& item)
{
    auto P_layout = std::make_unique<ParticleLayout>();
    double total_density = item.totalDensity();
    double layout_weight = item.weight();
    P_layout->setTotalParticleSurfaceDensity(total_density);
    P_layout->setWeight(layout_weight);
    return P_layout;
}

std::unique_ptr<IParticle> GUI::Transform::ToDomain::createIParticle(const SessionItem& item)
{
    std::unique_ptr<IParticle> P_particle;
    if (item.hasModelType<ParticleItem>()) {
        const auto& particle_item = dynamic_cast<const ParticleItem&>(item);
        P_particle = particle_item.createParticle();
    } else if (item.hasModelType<ParticleCoreShellItem>()) {
        const auto& particle_coreshell_item = dynamic_cast<const ParticleCoreShellItem&>(item);
        P_particle = particle_coreshell_item.createParticleCoreShell();
    } else if (item.hasModelType<ParticleCompositionItem>()) {
        const auto& particle_composition_item = dynamic_cast<const ParticleCompositionItem&>(item);
        P_particle = particle_composition_item.createParticleComposition();
    } else if (item.hasModelType<MesoCrystalItem>()) {
        const auto& mesocrystal_item = dynamic_cast<const MesoCrystalItem&>(item);
        P_particle = mesocrystal_item.createMesoCrystal();
    }
    return P_particle;
}
