//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DistributionEditor.cpp
//! @brief     Implements classes DistributionSelector and DistributionEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/Model/Group/DistributionItems.h"
#include "GUI/Model/Instrument/ItemWithDistribution.h"
#include "GUI/Model/Session/SessionModel.h"
#include "GUI/View/Instrument/DistributionForms.h"
#include "GUI/View/Instrument/EditDistributionDialog.h"
#include "GUI/View/PropertyEditor/GroupInfoBox.h"

#include <QComboBox>
#include <QFormLayout>
#include <QStackedLayout>
#include <utility>

namespace {

// some temporary functions to be able to assign the value of a distribution to an other

void assignBasic(const DistributionItem* src, DistributionItem* dst)
{
    dst->setNumberOfSamples(src->numberOfSamples());
    if (src->hasSigmaFactor())
        dst->setSigmaFactor(src->sigmaFactor());
}

void assignSymmetric(const SymmetricDistributionItem* src, SymmetricDistributionItem* dst,
                     const bool assignMean)
{
    assignBasic(src, dst);
    if (assignMean)
        dst->setMean(src->mean());
}

void assign(const DistributionNoneItem* src, DistributionNoneItem* dst, const bool assignMean)
{
    if (assignMean)
        dst->setMean(src->mean());
}

void assign(const DistributionGateItem* src, DistributionGateItem* dst)
{
    assignBasic(src, dst);
    dst->setRange(src->minimum(), src->maximum());
}

void assign(const DistributionLorentzItem* src, DistributionLorentzItem* dst, const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setHWHM(src->hwhm());
}

void assign(const DistributionGaussianItem* src, DistributionGaussianItem* dst,
            const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setStandardDeviation(src->standardDeviation());
}

void assign(const DistributionLogNormalItem* src, DistributionLogNormalItem* dst)
{
    assignBasic(src, dst);
    dst->setMedian(src->median());
    dst->setScaleParameter(src->scaleParameter());
}

void assign(const DistributionCosineItem* src, DistributionCosineItem* dst, const bool assignMean)
{
    assignSymmetric(src, dst, assignMean);
    dst->setSigma(src->sigma());
}

void assign(const DistributionTrapezoidItem* src, DistributionTrapezoidItem* dst)
{
    assignBasic(src, dst);
    dst->setCenter(src->center());
    dst->setLeftWidth(src->leftWidth());
    dst->setMiddleWidth(src->middleWidth());
    dst->setRightWidth(src->rightWidth());
}

void assign(const ItemWithDistribution* src, ItemWithDistribution* dst, const bool assignMean)
{
    dst->setDistributionType(src->distribution()->modelType());
    if (src->distribution()->hasModelType<DistributionNoneItem>())
        assign(dynamic_cast<const DistributionNoneItem*>(src->distribution()),
               dynamic_cast<DistributionNoneItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->hasModelType<DistributionGateItem>())
        assign(dynamic_cast<const DistributionGateItem*>(src->distribution()),
               dynamic_cast<DistributionGateItem*>(dst->distribution()));
    else if (src->distribution()->hasModelType<DistributionCosineItem>())
        assign(dynamic_cast<const DistributionCosineItem*>(src->distribution()),
               dynamic_cast<DistributionCosineItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->hasModelType<DistributionLorentzItem>())
        assign(dynamic_cast<const DistributionLorentzItem*>(src->distribution()),
               dynamic_cast<DistributionLorentzItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->hasModelType<DistributionGaussianItem>())
        assign(dynamic_cast<const DistributionGaussianItem*>(src->distribution()),
               dynamic_cast<DistributionGaussianItem*>(dst->distribution()), assignMean);
    else if (src->distribution()->hasModelType<DistributionLogNormalItem>())
        assign(dynamic_cast<const DistributionLogNormalItem*>(src->distribution()),
               dynamic_cast<DistributionLogNormalItem*>(dst->distribution()));
    else if (src->distribution()->hasModelType<DistributionTrapezoidItem>())
        assign(dynamic_cast<const DistributionTrapezoidItem*>(src->distribution()),
               dynamic_cast<DistributionTrapezoidItem*>(dst->distribution()));
    else
        ASSERT(false);
}

} // namespace

//==================================================================================================
// DistributionSelector
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

DistributionSelector::DistributionSelector(std::optional<MeanConfig> mean_config,
                                           GUI::ID::Distributions distributions, QWidget* parent,
                                           Qt::WindowFlags f)
    : QWidget(parent, f)
    , m_item(nullptr)
    , m_meanConfig(std::move(mean_config))
    , m_distributions(distributions)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_combo = new QComboBox(this);
    layout->addRow("Distribution:", m_combo);

    m_stack = new QStackedLayout();
    layout->addRow(m_stack);

    m_combo->addItem("Cosine", DistributionCosineItem::M_TYPE);
    auto* cosform = new CosineDistributionForm(m_meanConfig, this);
    connect(cosform, &CosineDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
    m_stack->addWidget(cosform);
    m_forms.emplace(DistributionCosineItem::M_TYPE, cosform);

    if (m_distributions == GUI::ID::Distributions::All) {
        m_combo->addItem("Gate", DistributionGateItem::M_TYPE);
        auto* gateform = new GateDistributionForm(this);
        connect(gateform, &GateDistributionForm::distributionChanged, this,
                &DistributionSelector::distributionChanged);
        m_stack->addWidget(gateform);
        m_forms.emplace(DistributionGateItem::M_TYPE, gateform);
    }

    m_combo->addItem("Gaussian", DistributionGaussianItem::M_TYPE);
    auto* gaussform = new GaussDistributionForm(m_meanConfig, this);
    connect(gaussform, &GaussDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
    m_stack->addWidget(gaussform);
    m_forms.emplace(DistributionGaussianItem::M_TYPE, gaussform);

    if (m_distributions == GUI::ID::Distributions::All) {
        m_combo->addItem("Log normal", DistributionLogNormalItem::M_TYPE);
        auto* lognormform = new LogNormalDistributionForm(this);
        connect(lognormform, &LogNormalDistributionForm::distributionChanged, this,
                &DistributionSelector::distributionChanged);
        m_stack->addWidget(lognormform);
        m_forms.emplace(DistributionLogNormalItem::M_TYPE, lognormform);
    }

    m_combo->addItem("Lorentz", DistributionLorentzItem::M_TYPE);
    auto* lorentzform = new LorentzDistributionForm(m_meanConfig, this);
    connect(lorentzform, &LorentzDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
    m_stack->addWidget(lorentzform);
    m_forms.emplace(DistributionLorentzItem::M_TYPE, lorentzform);

    m_combo->addItem("None", DistributionNoneItem::M_TYPE);
    auto* noneform = new NoneDistributionForm(m_meanConfig, this);
    connect(noneform, &NoneDistributionForm::distributionChanged, this,
            &DistributionSelector::distributionChanged);
    m_stack->addWidget(noneform);
    m_forms.emplace(DistributionNoneItem::M_TYPE, noneform);

    if (m_distributions == GUI::ID::Distributions::All) {
        m_combo->addItem("Trapezoid", DistributionTrapezoidItem::M_TYPE);
        auto* trapezoidform = new TrapezoidDistributionForm(this);
        connect(trapezoidform, &TrapezoidDistributionForm::distributionChanged, this,
                &DistributionSelector::distributionChanged);
        m_stack->addWidget(trapezoidform);
        m_forms.emplace(DistributionTrapezoidItem::M_TYPE, trapezoidform);
    }

    connect(m_combo, qOverload<int>(&QComboBox::currentIndexChanged), this,
            &DistributionSelector::onDistributionSelected);
}

void DistributionSelector::setItem(ItemWithDistribution* item)
{
    m_item = item;
    refresh();
}

ItemWithDistribution* DistributionSelector::item() const
{
    return m_item;
}

const std::optional<MeanConfig>& DistributionSelector::meanConfig() const
{
    return m_meanConfig;
}

GUI::ID::Distributions DistributionSelector::distributions() const
{
    return m_distributions;
}

void DistributionSelector::clearItem()
{
    setItem(nullptr);
}

void DistributionSelector::refresh()
{
    setEnabled(m_item);
    if (m_item) {
        int idx = m_combo->findData(m_item->distribution()->modelType());
        ASSERT(idx != -1);
        if (idx != m_combo->currentIndex())
            m_combo->setCurrentIndex(idx);
        else
            onDistributionSelected(idx); // trigger update manually
    }
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void DistributionSelector::onDistributionSelected(int index)
{
    DistributionForm* form = m_forms[m_combo->itemData(index).toString()];
    form->setupDistribution(m_item);
    m_stack->setCurrentWidget(form);
    emit distributionChanged();
}

//==================================================================================================
// DistributionEditor
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

DistributionEditor::DistributionEditor(const QString& title,
                                       const std::optional<MeanConfig>& mean_config,
                                       GUI::ID::Distributions distributions, QWidget* parent,
                                       Qt::WindowFlags f)
    : QWidget(parent, f), m_title(title)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new GroupInfoBox(title, this);
    auto* glayout = new QVBoxLayout(gbox);
    m_selector = new DistributionSelector(mean_config, distributions, gbox);
    glayout->addWidget(m_selector);
    layout->addWidget(gbox);

    connect(m_selector, &DistributionSelector::distributionChanged, this,
            &DistributionEditor::distributionChanged);
    connect(gbox, &GroupInfoBox::clicked, this, &DistributionEditor::showDialog);
}

void DistributionEditor::setItem(ItemWithDistribution* item)
{
    m_selector->setItem(item);
}

void DistributionEditor::clearItem()
{
    m_selector->clearItem();
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void DistributionEditor::showDialog()
{
    showEditDistributionDialog(m_selector, m_title);
}

//--------------------------------------------------------------------------------------------------
// free functions
//--------------------------------------------------------------------------------------------------

void showEditDistributionDialog(DistributionSelector* selector, const QString& title)
{
    ItemWithDistribution* item = selector->item();
    if (!item)
        return;

    SessionModel dummy("dummy");
    ItemWithDistribution* distr_item =
        dynamic_cast<ItemWithDistribution*>(dummy.copy(dynamic_cast<SessionItem*>(item)));
    EditDistributionDialog dialog(distr_item, title, selector->meanConfig(),
                                  selector->distributions(), selector);
    if (dialog.exec() == QDialog::Accepted) {
        assign(distr_item, item, selector->meanConfig().has_value());
        selector->refresh();
    }
}
