//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/VectorEditor.cpp
//! @brief     Implements class VectorEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/VectorEditor.h"
#include "GUI/Model/Types/VectorItem.h"

#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>

#include <limits>

//==================================================================================================
// VectorEditor
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

VectorEditor::VectorEditor(const QString& title, const QString& tooltip, QWidget* parent,
                           Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new QGroupBox(title, this);
    layout->addWidget(gbox);
    auto* form = new QFormLayout(gbox);

    m_xSpinBox = new QDoubleSpinBox(gbox);
    m_xSpinBox->setToolTip(tooltip.arg("x coordinate"));
    m_xSpinBox->setDecimals(4);
    m_xSpinBox->setRange(std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max());
    form->addRow("X:", m_xSpinBox);

    m_ySpinBox = new QDoubleSpinBox(gbox);
    m_ySpinBox->setToolTip(tooltip.arg("y coordinate"));
    m_ySpinBox->setDecimals(4);
    m_ySpinBox->setRange(std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max());
    form->addRow("Y:", m_ySpinBox);

    m_zSpinBox = new QDoubleSpinBox(gbox);
    m_zSpinBox->setToolTip(tooltip.arg("z coordinate"));
    m_zSpinBox->setDecimals(4);
    m_zSpinBox->setRange(std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max());
    form->addRow("Z:", m_zSpinBox);

    connect(m_xSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &VectorEditor::onXValueChanged);
    connect(m_ySpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &VectorEditor::onYValueChanged);
    connect(m_zSpinBox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &VectorEditor::onZValueChanged);
}

void VectorEditor::setItem(VectorItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item) {
        m_xSpinBox->setValue(m_item->x());
        m_ySpinBox->setValue(m_item->y());
        m_zSpinBox->setValue(m_item->z());
    } else {
        m_xSpinBox->setValue(0);
        m_ySpinBox->setValue(0);
        m_zSpinBox->setValue(0);
    }
}

VectorItem* VectorEditor::item() const
{
    return m_item;
}

void VectorEditor::clearItem()
{
    setItem(nullptr);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void VectorEditor::onXValueChanged(double x)
{
    if (m_item) {
        m_item->setX(x);
        emit vectorChanged();
    }
}

void VectorEditor::onYValueChanged(double y)
{
    if (m_item) {
        m_item->setY(y);
        emit vectorChanged();
    }
}

void VectorEditor::onZValueChanged(double z)
{
    if (m_item) {
        m_item->setZ(z);
        emit vectorChanged();
    }
}
