//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/ParticleLayoutForm.cpp
//! @brief     Implements class ParticleLayoutForm
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/ParticleLayoutForm.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/ItemWithParticles.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Util/ActionFactory.h"
#include "GUI/View/Edit/DoubleSpinBox.h"
#include "GUI/View/SampleDesigner/InterferenceForm.h"
#include "GUI/View/SampleDesigner/LayerEditorUtils.h"
#include "GUI/View/SampleDesigner/LayerForm.h"
#include "GUI/View/SampleDesigner/SampleEditorController.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"

#include <QAction>
#include <QPushButton>

ParticleLayoutForm::ParticleLayoutForm(LayerForm* parent, ParticleLayoutItem* layoutItem,
                                       SampleEditorController* ec)
    : QGroupBox(parent), m_layoutItem(layoutItem), m_ec(ec)
{
    setTitle(m_layoutItem->itemName());

    FormLayouter layouter(this, ec);
    layouter.setContentsMargins(30, 6, 0, 0);
    int rowOfTotalDensity = layouter.addValue(m_layoutItem->totalDensity());
    m_totalDensitySpinBox =
        layouter.widgetAt<DoubleSpinBox*>(rowOfTotalDensity, QFormLayout::FieldRole);
    ASSERT(m_totalDensitySpinBox);
    layouter.addValue(m_layoutItem->weight());
    layouter.addRow(new InterferenceForm(this, layoutItem, ec));

    for (auto* particle : m_layoutItem->particles())
        layouter.addRow(
            LayerEditorUtils::createWidgetForItemWithParticles(this, particle, true, ec));

    auto* btn = LayerEditorUtils::createAddParticleButton(
        this, [=](const QString& cls) { ec->addParticle(layoutItem, cls); });
    m_structureEditingWidgets << btn;
    layouter.addStructureEditingRow(btn);

    m_removeAction = ActionFactory::createRemoveAction(
        this, "particle layout", [=] { ec->removeLayout(parent, layoutItem); });

    auto* showInRealSpaceAction = ActionFactory::createShowInRealSpaceAction(
        this, "particle layout", [=] { ec->requestViewInRealSpace(layoutItem); });

    auto* collapser = GroupBoxCollapser::installIntoGroupBox(this);
    collapser->addAction(showInRealSpaceAction);
    collapser->addAction(m_removeAction);
    m_layout = layouter.layout();

    updateDensityEnabling();
}

void ParticleLayoutForm::enableStructureEditing(bool b)
{
    m_removeAction->setVisible(b);
    for (auto* w : m_structureEditingWidgets)
        w->setVisible(b);
}

ParticleLayoutItem* ParticleLayoutForm::layoutItem() const
{
    return m_layoutItem;
}

void ParticleLayoutForm::onParticleAdded(SessionItem* new_item)
{
    if (auto* p = dynamic_cast<ItemWithParticles*>(new_item)) {
        int index = m_layoutItem->particles().indexOf(p);
        const int rowInLayout =
            m_layout->rowCount() - 1 - (m_layoutItem->particles().size() - 1) + index; // -1: btn

        m_layout->insertRow(
            rowInLayout, LayerEditorUtils::createWidgetForItemWithParticles(this, p, true, m_ec));
    }
}

void ParticleLayoutForm::onAboutToRemoveParticle(ItemWithParticles* item)
{
    int index = m_layoutItem->particles().indexOf(item);
    const int rowInLayout =
        m_layout->rowCount() - m_layoutItem->particles().size() - 1 + index; // -1: btn

    m_layout->removeRow(rowInLayout);
}

void ParticleLayoutForm::updateDensityEnabling()
{
    const auto* currentInterference = m_layoutItem->interference().currentItem();
    const bool enableTotalDensityInParticleLayout =
        currentInterference == nullptr
        || (currentInterference->modelType() != Interference2DLatticeItem::M_TYPE
            && currentInterference->modelType() != Interference2DParaCrystalItem::M_TYPE
            && currentInterference->modelType() != InterferenceFinite2DLatticeItem::M_TYPE
            && currentInterference->modelType() != InterferenceHardDiskItem::M_TYPE);

    m_totalDensitySpinBox->setEnabled(enableTotalDensityInParticleLayout);
}

void ParticleLayoutForm::updateDensityValue()
{
    m_totalDensitySpinBox->updateValue();
}
