//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Toplevel/SessionModelView.cpp
//! @brief     Implements class SessionModelView
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Toplevel/SessionModelView.h"
#include "GUI/Model/Data/RealDataModel.h"
#include "GUI/Model/Instrument/InstrumentModel.h"
#include "GUI/Model/Job/JobModel.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Session/SessionDecorationModel.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/View/Main/ProjectManager.h"
#include "GUI/View/Tool/StyleUtils.h"
#include <QApplication>
#include <QStyledItemDelegate>
#include <QToolBar>
#include <QToolButton>
#include <QTreeView>
#include <QVBoxLayout>

namespace {
//! The SessionModelDelegate class presents the content of SessionModel items in
//! standard QTreeView. Extents base QItemDelegate with possibility to show/edit
//! our custom QVariant's.

class SessionModelDelegate : public QStyledItemDelegate {
public:
    SessionModelDelegate(QObject* parent) : QStyledItemDelegate(parent) {}

    void paint(QPainter* painter, const QStyleOptionViewItem& option,
               const QModelIndex& index) const override
    {
        const bool isMaterialProperty =
            index.data(SessionFlags::CustomEditorRole) == SessionItem::EDITOR_TYPE_MATERIAL
            && index.column() == SessionFlags::ITEM_VALUE;

        if (isMaterialProperty) {
            const auto* materialItem = static_cast<SessionItem*>(index.internalPointer());
            const auto* itemWithMaterial = dynamic_cast<ItemWithMaterial*>(materialItem->parent());
            ASSERT(itemWithMaterial);
            paintCustomLabel(painter, option, index, itemWithMaterial->materialName());
        } else if (index.data().canConvert<ComboProperty>())
            paintCustomLabel(painter, option, index, index.data().value<ComboProperty>().label());
        else
            QStyledItemDelegate::paint(painter, option, index);
    }

private:
    void paintCustomLabel(QPainter* painter, const QStyleOptionViewItem& option,
                          const QModelIndex& index, const QString& text) const
    {
        QStyleOptionViewItem opt = option;
        initStyleOption(&opt, index); // calling original method to take into accounts colors etc
        opt.text = displayText(text, option.locale); // by overriding text with ours
        const QWidget* widget = opt.widget;
        QStyle* style = widget ? widget->style() : QApplication::style();
        style->drawControl(QStyle::CE_ItemViewItem, &opt, painter, widget);
    }
};

} // namespace

SessionModelView::SessionModelView(QWidget* parent, ProjectDocument* document) : QWidget(parent)
{
    auto* layout = new QVBoxLayout(this);
    layout->setMargin(0);

    auto* tabs = new QTabWidget(this);
    auto* m_expandButton = new QToolButton(this);
    m_expandButton->setText("Expand tree");
    m_expandButton->setIcon(QIcon(":/images/toolbar_expand_collapse_tree.svg"));
    m_expandButton->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    connect(m_expandButton, &QToolButton::clicked, [=]() {
        auto* tree = static_cast<QTreeView*>(tabs->widget(tabs->currentIndex()));
        tree->expandAll();
        tree->resizeColumnToContents(0);
        tree->resizeColumnToContents(1);
    });

    auto* m_collapseButton = new QToolButton(this);
    m_collapseButton->setText("Collapse tree");
    m_collapseButton->setIcon(QIcon(":/images/toolbar_expand_collapse_tree.svg"));
    m_collapseButton->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    connect(m_collapseButton, &QToolButton::clicked, [=]() {
        auto* tree = static_cast<QTreeView*>(tabs->widget(tabs->currentIndex()));
        tree->collapseAll();
    });

    auto* toolBar = new QToolBar;
    toolBar->addWidget(m_expandButton);
    toolBar->addWidget(m_collapseButton);
    layout->addWidget(toolBar);
    layout->addWidget(tabs);

    QList<SessionModel*> models{document->instrumentModel(), document->sampleModel(),
                                document->realDataModel(), document->materialModel(),
                                document->jobModel()};

    for (auto* model : models) {
        auto* treeView = new QTreeView(this);
        GUI::Util::Style::setPropertyStyle(treeView);
        treeView->setItemDelegate(new SessionModelDelegate(this));
        treeView->setModel(new SessionDecorationModel(treeView, model));
        treeView->setEditTriggers(QAbstractItemView::NoEditTriggers); // read-only!
        tabs->addTab(treeView, model->getModelTag());
    }
}
