//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Resample/Processed/ProcessedSample.h
//! @brief     Defines class ProcessedSample.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_RESAMPLE_PROCESSED_PROCESSEDSAMPLE_H
#define BORNAGAIN_RESAMPLE_PROCESSED_PROCESSEDSAMPLE_H

#include "Base/Vector/Vectors3D.h"
#include "Resample/Slice/SliceStack.h"
#include <memory>
#include <vector>

class IFlux;
class Material;
class MultiLayer;
class ProcessedLayout;
class SimulationOptions;

using Fluxes = std::vector<std::unique_ptr<const IFlux>>;

//! Data structure that contains all the necessary data for scattering calculations.
//!
//! If the usage of average materials is requested, layers and particles are sliced into multiple
//! slices and the average material is calculated for each slice.

class ProcessedSample {
public:
    //! Factory method that wraps the private constructor.
    static ProcessedSample make(const MultiLayer& sample, const SimulationOptions& options,
                                bool forcePolarized = false);
    ProcessedSample(const ProcessedSample&) = delete;
    ~ProcessedSample();

    size_t numberOfSlices() const;
    const SliceStack& averageSlices() const;
    const Slice& avgeSlice(size_t i) const;
    const std::vector<ProcessedLayout>& layouts() const;
    double sliceTopZ(size_t i) const;
    double sliceBottomZ(size_t i) const;
    bool containsMagneticMaterial() const;
    bool polarizing() const; //!< Contains magnetic material, or nonzero magnetic field
    bool hasRoughness() const;
    const MultiLayer& sample() const { return m_sample; }

    Fluxes fluxesIn(const R3& k) const;
    Fluxes fluxesOut(const R3& k) const;

    //! Fourier transform of the correlation function of roughnesses between the interfaces
    double crossCorrSpectralFun(R3 kvec, size_t j, size_t k) const;

private:
    ProcessedSample(const MultiLayer& sample, bool polarized,
                    std::vector<ProcessedLayout>&& layouts, const SliceStack& refined_stack);
    const MultiLayer& m_sample;
    const bool m_polarized;
    std::vector<ProcessedLayout> m_layouts; // const forbidden by &&-c'tor < needed by tests
    const SliceStack m_stack;
};

#endif // BORNAGAIN_RESAMPLE_PROCESSED_PROCESSEDSAMPLE_H
#endif // USER_API
