//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Aggregate/ParticleLayout.h
//! @brief     Defines class ParticleLayout.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H
#define BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H

#include "Base/Types/SafePointerVector.h"
#include "Sample/Scattering/ISampleNode.h"
#include "Sample/Scattering/Rotations.h"
#include <memory>

class IAbstractParticle;
class IInterference;
class IParticle;

//! Decorator class that adds particles to ISampleNode objects.
//! @ingroup samples

class ParticleLayout : public ISampleNode {
public:
    inline static const std::string class_name = "ParticleLayout";
    std::string className() const final { return class_name; }

    ParticleLayout();
    ParticleLayout(const IAbstractParticle& particle, double abundance = -1.0);
    ~ParticleLayout() override;

    ParticleLayout* clone() const override;

    void accept(INodeVisitor* visitor) const override { visitor->visit(this); }

    void addParticle(const IAbstractParticle& particle, double abundance = -1.0, R3 position = {},
                     const IRotation& rotation = IdentityRotation());

    SafePointerVector<IParticle> particles() const;

    const IInterference* interferenceFunction() const;

    double getTotalAbundance() const;

    void setInterference(const IInterference& interparticle);

    double totalParticleSurfaceDensity() const;
    void setTotalParticleSurfaceDensity(double particle_density);

    std::vector<const INode*> getChildren() const override;

    //! Returns the relative weight of this layout
    double weight() const { return m_weight; }

    //! Sets the relative weight of this layout
    void setWeight(double weight) { m_weight = weight; }

    double weightedParticleSurfaceDensity() const;

private:
    void addAndRegisterAbstractParticle(IAbstractParticle* child);
    void setAndRegisterInterference(IInterference* child);

    double m_weight;
    double m_total_particle_density;
    SafePointerVector<IAbstractParticle> m_particles; //!< Vector of particle types
    std::unique_ptr<IInterference> m_interparticle;
};

#endif // BORNAGAIN_SAMPLE_AGGREGATE_PARTICLELAYOUT_H
