//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Material/MagneticMaterialImpl.h
//! @brief     Defines magnetic material base implementation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_SAMPLE_MATERIAL_MAGNETICMATERIALIMPL_H
#define BORNAGAIN_SAMPLE_MATERIAL_MAGNETICMATERIALIMPL_H

#include "Sample/Material/BaseMaterialImpl.h"

class Transform3D;
class WavevectorInfo;

//! Basic implementation for magnetized material.
//! Inherited by RefractiveMaterialImpl and MaterialBySLDImpl.
//! Incorporates data and methods required to handle material magnetization.
//! @ingroup materials

class MagneticMaterialImpl : public BaseMaterialImpl {
public:
    //! Constructs basic material with name and magnetization
    MagneticMaterialImpl(const std::string& name, R3 magnetization);

    ~MagneticMaterialImpl() override = default;

    //! Returns pointer to a copy of material
    MagneticMaterialImpl* clone() const override = 0;

    //! Constructs a material with inverted magnetization
    MagneticMaterialImpl* inverted() const override;

    //! Indicates whether the interaction with the material is scalar.
    //! This means that different polarization states will be diffracted equally
    bool isScalarMaterial() const override;

    bool isMagneticMaterial() const override;

    //! Returns the magnetization (in A/m)
    R3 magnetization() const override;

    //! Returns (\f$ \pi/\lambda^2 \f$ - sld) matrix with magnetization corrections
    Eigen::Matrix2cd polarizedSubtrSLD(const WavevectorInfo& wavevectors) const override;

    MagneticMaterialImpl* rotatedMaterial(const Transform3D& transform) const override;

private:
    void setMagnetization(R3 magnetization) { m_magnetization = magnetization; }

    R3 m_magnetization; //!< magnetization
};

#endif // BORNAGAIN_SAMPLE_MATERIAL_MAGNETICMATERIALIMPL_H
#endif // USER_API
