//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/Crystal.h
//! @brief     Defines class Crystal.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SAMPLE_PARTICLE_CRYSTAL_H
#define BORNAGAIN_SAMPLE_PARTICLE_CRYSTAL_H

#include "Sample/Lattice/Lattice3D.h"
#include "Sample/Scattering/ISampleNode.h"

class IFormFactor;
class IRotation;
class Admixtures;

//! A crystal structure, defined by a Bravais lattice, a basis, and a position variance.
//!
//! The basis is either a Particle or a ParticleComposition.
//!
//! Computations are delegated to class FormFactorCrystal.
//!
//! Used in MesoCrystal, where it is given an outer shape.
//!
//! @ingroup samples

class Crystal : public ISampleNode {
public:
    inline static const std::string class_name = "Crystal";
    std::string className() const final { return class_name; }

    Crystal(const IParticle& basis, const Lattice3D& lattice, double position_variance = 0);
    ~Crystal() override;

    Crystal* clone() const override;

    void accept(INodeVisitor* visitor) const override { visitor->visit(this); }

    IFormFactor* createTotalFormFactor(const IFormFactor& meso_crystal_form_factor,
                                       const IRotation* rotation, const R3& translation) const;

    Admixtures admixtures() const;

    Lattice3D transformedLattice(const IRotation* rotation = nullptr) const;

    std::vector<const INode*> getChildren() const override;

private:
    Crystal(IParticle* p_basis, const Lattice3D& lattice, double position_variance = 0);

    Lattice3D m_lattice;
    std::unique_ptr<IParticle> m_basis;
    const double m_position_variance;
};

#endif // BORNAGAIN_SAMPLE_PARTICLE_CRYSTAL_H
