#include "GUI/Model/Data/ApplicationModels.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/View/Realspace/Particle3DContainer.h"
#include "GUI/View/Realspace/RealSpaceBuilder.h"
#include "GUI/View/Realspace/RealSpaceBuilderUtils.h"
#include "GUI/View/Realspace/TransformTo3D.h"
#include "GUI/ba3d/model/model.h"
#include "Sample/Particle/Particle.h"
#include "Tests/GTestWrapper/google_test.h"
#include <QObject>

class TestRealSpaceBuilderUtils : public ::testing::Test {
};

TEST_F(TestRealSpaceBuilderUtils, RealSpaceModelandParticle)
{
    GUI::RealSpace::Model realSpaceModel;

    auto cylinder3D = std::make_unique<GUI::RealSpace::Particles::Cylinder>(5, 10);
    QVector3D lattice_position1(0, 0, 0);
    cylinder3D->addTranslation(lattice_position1);
    realSpaceModel.add(cylinder3D.release());

    // Causes a failure as one cannot add an already added particle to the model
    //    QVector3D lattice_position2(10,10,10);
    //    cylinder3D->addTranslation(lattice_position2);
    //    realSpaceModel.add(cylinder3D.release());
}

TEST_F(TestRealSpaceBuilderUtils, computeCumulativeAbundances)
{
    SampleModel sampleModel;
    MaterialModel materialModel;
    auto* layout = sampleModel.insertItem<ParticleLayoutItem>();

    auto* particle1 = sampleModel.insertItem<ParticleItem>();
    layout->addParticle(particle1);
    EXPECT_EQ(particle1->parent(), layout);

    auto* particle2 = sampleModel.insertItem<ParticleItem>();
    EXPECT_EQ(particle2->parent(), sampleModel.rootItem());

    layout->addParticle(particle2);
    EXPECT_EQ(particle2->parent(), layout);

    particle1->setAbundance(8.0);
    particle2->setAbundance(2.0);

    EXPECT_EQ(
        GUI::RealSpace::BuilderUtils(&materialModel).computeCumulativeAbundances(*layout).last(),
        10.0);
}

TEST_F(TestRealSpaceBuilderUtils, Particle3DContainer)
{
    Particle3DContainer p1;

    EXPECT_EQ(p1.containerSize(), 0u);
    EXPECT_EQ(p1.cumulativeAbundance(), 0);
    EXPECT_EQ(p1.particleType(), "");

    // Create 3D cylinder using corresponding constructor GUI::RealSpace::Particles::Cylinder()
    auto cylinder3D = std::make_unique<GUI::RealSpace::Particles::Cylinder>(5, 10);

    // Associate it to a Particle3DContainer object
    p1.addParticle(cylinder3D.release(), false);
    p1.setCumulativeAbundance(1);
    p1.setParticleType("Particle");

    EXPECT_EQ(p1.containerSize(), 1u);
    EXPECT_EQ(p1.cumulativeAbundance(), 1);
    EXPECT_EQ(p1.particleType(), "Particle");
    EXPECT_FALSE(p1.particle3DBlend(0u));

    GUI::RealSpace::Model realSpaceModel;

    // Test copy constructor
    // If copy constructor (DEEP copy) is not implemented then p2/p3 is basically
    // the same as p1 and cannot be added to the GUI::RealSpace::Model once p1 is added
    Particle3DContainer p2(p1);

    EXPECT_EQ(p2.containerSize(), 1u);
    EXPECT_EQ(p2.cumulativeAbundance(), 1);
    EXPECT_EQ(p2.particleType(), "Particle");
    EXPECT_FALSE(p2.particle3DBlend(0u));

    Particle3DContainer p3 = {p1};

    EXPECT_EQ(p3.containerSize(), 1u);
    EXPECT_EQ(p3.cumulativeAbundance(), 1);
    EXPECT_EQ(p3.particleType(), "Particle");
    EXPECT_FALSE(p3.particle3DBlend(0u));

    // Test copy assignment operator
    // If assignment operator (DEEP) is not implemented then p4 is basically
    // the same as p1 and cannot be added to the GUI::RealSpace::Model once p1 is added
    Particle3DContainer p4;
    p4 = p1;

    EXPECT_EQ(p4.containerSize(), 1u);
    EXPECT_EQ(p4.cumulativeAbundance(), 1);
    EXPECT_EQ(p4.particleType(), "Particle");
    EXPECT_FALSE(p4.particle3DBlend(0u));

    // Add particle to GUI::RealSpace::Model by first creating a unique instance of it and then
    // releasing (transfer ownership) to the model which deletes the new particle instance later on
    auto p1_unique = p1.createParticle(0);
    auto p2_unique = p2.createParticle(0);
    auto p3_unique = p3.createParticle(0);
    auto p4_unique = p4.createParticle(0);

    realSpaceModel.add(p1_unique.release());
    realSpaceModel.add(p2_unique.release());
    realSpaceModel.add(p3_unique.release());
    realSpaceModel.add(p4_unique.release());

    // Test move constructor
    Particle3DContainer p5(std::move(p4));
    EXPECT_EQ(p5.containerSize(), 1u);
    EXPECT_EQ(p5.cumulativeAbundance(), 1);
    EXPECT_EQ(p5.particleType(), "Particle");
    EXPECT_FALSE(p5.particle3DBlend(0u));

    auto p5_unique = p5.createParticle(0);
    realSpaceModel.add(p5_unique.release());

    Particle3DContainer p6 = Particle3DContainer(std::move(p5));
    EXPECT_EQ(p6.containerSize(), 1u);
    EXPECT_EQ(p6.cumulativeAbundance(), 1);
    EXPECT_EQ(p6.particleType(), "Particle");
    EXPECT_FALSE(p6.particle3DBlend(0u));

    auto p6_unique = p6.createParticle(0);
    realSpaceModel.add(p6_unique.release());

    // Test move assignment operator
    Particle3DContainer p7;
    p7 = std::move(p6);
    EXPECT_EQ(p7.containerSize(), 1u);
    EXPECT_EQ(p7.cumulativeAbundance(), 1);
    EXPECT_EQ(p7.particleType(), "Particle");
    EXPECT_FALSE(p7.particle3DBlend(0u));

    auto p7_unique = p7.createParticle(0);
    realSpaceModel.add(p7_unique.release());

    // The normal raw instances p1, p2 etc. get deleted by the destructor of Particle3DContainer
}

TEST_F(TestRealSpaceBuilderUtils, singleParticle3DContainer)
{
    ApplicationModels models;
    SampleModel* sampleModel = models.sampleModel();

    auto* particleItem = sampleModel->insertItem<ParticleItem>();
    particleItem->setMaterial(models.materialModel()->defaultMaterial());
    EXPECT_EQ(particleItem->abundance(), 1.0);
    EXPECT_TRUE(particleItem->formFactor()->hasModelType<CylinderItem>());

    particleItem->setAbundance(8.0);
    EXPECT_EQ(particleItem->abundance(), 8.0);
    particleItem->setFormFactorType<BoxItem>();
    EXPECT_TRUE(particleItem->formFactor()->hasModelType<BoxItem>());

    // Create a 3D particle from particleItem and associate it to a Particle3DContainer object
    const auto* pItem = dynamic_cast<const ParticleItem*>(particleItem);
    auto particle = pItem->createParticle();
    auto singleParticle3DContainer = GUI::RealSpace::BuilderUtils(models.materialModel())
                                         .singleParticle3DContainer(*particle, 8);

    EXPECT_EQ(singleParticle3DContainer.containerSize(), 1u);
    EXPECT_EQ(singleParticle3DContainer.cumulativeAbundance(), 1);
    EXPECT_EQ(singleParticle3DContainer.particleType(), "Particle");
    EXPECT_FALSE(singleParticle3DContainer.particle3DBlend(0u));
}

TEST_F(TestRealSpaceBuilderUtils, particle3DContainerVector)
{
    ApplicationModels models;
    SampleModel* sampleModel = models.sampleModel();

    auto* layout = sampleModel->insertItem<ParticleLayoutItem>();
    auto* particle1 = sampleModel->insertItem<ParticleItem>();
    auto* particle2 = sampleModel->insertItem<ParticleItem>();
    auto* particle3 = sampleModel->insertItem<ParticleItem>();
    particle1->setMaterial(models.materialModel()->defaultMaterial());
    particle2->setMaterial(models.materialModel()->defaultMaterial());
    particle3->setMaterial(models.materialModel()->defaultMaterial());

    layout->addParticle(particle1);
    layout->addParticle(particle2);
    layout->addParticle(particle3);

    particle1->setAbundance(5.0);
    particle2->setAbundance(3.0);
    particle3->setAbundance(2.0);

    double total_abundance = GUI::RealSpace::BuilderUtils(models.materialModel())
                                 .computeCumulativeAbundances(*layout)
                                 .last();
    EXPECT_EQ(total_abundance, 10.0);

    particle1->setFormFactorType<BoxItem>();
    particle2->setFormFactorType<ConeItem>();
    particle3->setFormFactorType<PyramidItem>();

    auto particle3DContainer_vector =
        GUI::RealSpace::BuilderUtils(models.materialModel()).particle3DContainerVector(*layout);

    EXPECT_EQ(particle3DContainer_vector.size(), static_cast<size_t>(3));

    EXPECT_EQ(particle3DContainer_vector.at(0).containerSize(), 1u);
    EXPECT_EQ(particle3DContainer_vector.at(0).cumulativeAbundance(), 0.5);
    EXPECT_EQ(particle3DContainer_vector.at(0).particleType(), "Particle");

    EXPECT_EQ(particle3DContainer_vector.at(1).containerSize(), 1u);
    EXPECT_EQ(particle3DContainer_vector.at(1).cumulativeAbundance(), 0.8);
    EXPECT_EQ(particle3DContainer_vector.at(1).particleType(), "Particle");

    EXPECT_EQ(particle3DContainer_vector.at(2).containerSize(), 1u);
    EXPECT_EQ(particle3DContainer_vector.at(2).cumulativeAbundance(), 1.0);
    EXPECT_EQ(particle3DContainer_vector.at(2).particleType(), "Particle");
}
